//
#include <osg/PositionAttitudeTransform>
#include <osg/Group>
#include <osg/Node>
#include <osgDB/ReadFile>
#include <osgDB/FileUtils>
#include <osgProducer/Viewer>

#include <osgText/Font>
#include <osgText/Text>
#include <osg/MatrixTransform>
#include <osg/Geode>
#include <osg/Projection>
#include <osg/ShapeDrawable>
#include <osg/Geometry>
#include <osg/PolygonStipple>

#include <Producer/RenderSurface>
#include <Producer/KeyboardMouse>
#include <Producer/Trackball>
#include <osg/io_utils>
#include <osg/Math>

#include <sstream>
#include <map>

char tempstring[] = "                                                                                 ";

#define MAXSEARCHPATHS  10   // maximum number of paths added to search list for terrain textures

#include "KeyboardHandler.h"

extern bool globalStartPolygon;
extern bool globalEndPolygon;
extern bool globalAddNewPath;
extern std::string globalOSGTerrainName;
extern std::string globalSavedDir[MAXSEARCHPATHS];
extern std::string globalXMLFilename;
extern int globalNumberSearchPaths;

enum terrainType  {TWOINCHWASHBOARD=0 , SIXINCHWASHBOARD=1, RADIALWASHBOARD=2, 
                   BELGIANBLOCK=3, EMBEDDEDROCK=4, CROSSCOUNTRY=5, GRAVELROAD=6,
                   PAVEDROAD=7, NOTTRAVERSABLE=8};

#define		BBSCALE 1.0
osg::BoundingBox bb ;
int boxScaleFactor = 1;


#define     XMOUSEMIN	-1.0
#define		XMOUSEMAX	1.0
#define		YMOUSEMIN	-1.0
#define		YMOUSEMAX	1.0

#define		MAXVERTICES		100   // max number of vertices in a user-defined polygon

osg::Geode* myGeode = new osg::Geode();

bool bInitFile = false;
std::ofstream stream;
int vertexCount = 0;
int polygonCount = 0;

float xoff, yoff, zoff;
float xdiff, ydiff, zdiff;

int selectedTerrainType = TWOINCHWASHBOARD;

std::multimap<int, osg::Vec3f> mapPolygonPoints;
std::multimap<int, int> mapPolygonColor;

// Viewer instance to handle scene rendering:
   //osgProducer::Viewer viewer;
   // Global OSG constructs
Producer::CameraConfig* myCameraConfig = new Producer::CameraConfig();
Producer::Camera* tankFollowerCamera = new Producer::Camera();
Producer::Camera* mainCamera = new Producer::Camera();

void ExitApp(void)
{
	stream << "</MGXML>"  << std::endl;
	//viewer.done();
}
void TerrainSelectTWOINCHWASHBOARD(void)
{
	selectedTerrainType = TWOINCHWASHBOARD;
}

void TerrainSelectSIXINCHWASHBOARD(void)
{
	selectedTerrainType = SIXINCHWASHBOARD;
}

void TerrainSelectRADIALWASHBOARD(void)
{
	selectedTerrainType = RADIALWASHBOARD;
}

void TerrainSelectBELGIANBLOCK(void)
{
	selectedTerrainType = BELGIANBLOCK;
}

void TerrainSelectEMBEDDEDROCK(void)
{
	selectedTerrainType = EMBEDDEDROCK;
}

void TerrainSelectCROSSCOUNTRY(void)
{
	selectedTerrainType = CROSSCOUNTRY;
}

void TerrainSelectGRAVELROAD(void)
{
	selectedTerrainType = GRAVELROAD;
}

void TerrainSelectPAVEDROAD(void)
{
	selectedTerrainType = PAVEDROAD;
}

void TerrainSelectNOTTRAVERSABLE(void)
{
	selectedTerrainType = NOTTRAVERSABLE;
}

void DrawBox(float xVal, float yVal)
{
	//mg 
   // create LINES
    
        // create Geometry object to store all the vetices and lines primtive.
        osg::Geometry* linesGeom = new osg::Geometry();
        
        // this time we'll preallocate the vertex array to the size we
        // need and then simple set them as array elements, 8 points
        // makes 4 line segments.
        osg::Vec3Array* vertices = new osg::Vec3Array(8);
		float zOff = 0.0;
		float halfSize = 2.0;

        (*vertices)[0].set(xVal-halfSize*boxScaleFactor, yVal-halfSize*boxScaleFactor, zOff);
        (*vertices)[1].set(xVal+halfSize*boxScaleFactor, yVal-halfSize*boxScaleFactor, zOff);
        (*vertices)[2].set(xVal+halfSize*boxScaleFactor, yVal-halfSize*boxScaleFactor, zOff);
        (*vertices)[3].set(xVal+halfSize*boxScaleFactor, yVal+halfSize*boxScaleFactor, zOff);
		(*vertices)[4].set(xVal+halfSize*boxScaleFactor, yVal+halfSize*boxScaleFactor, zOff);
        (*vertices)[5].set(xVal-halfSize*boxScaleFactor, yVal+halfSize*boxScaleFactor, zOff);
        (*vertices)[6].set(xVal-halfSize*boxScaleFactor, yVal+halfSize*boxScaleFactor, zOff);
        (*vertices)[7].set(xVal-halfSize*boxScaleFactor, yVal-halfSize*boxScaleFactor, zOff);
        
        // pass the created vertex array to the points geometry object.
        linesGeom->setVertexArray(vertices);
        
        // set the colors as before, plus using the aobve
        osg::Vec4Array* mycolors = new osg::Vec4Array;
        mycolors->push_back(osg::Vec4(1.0f,0.0f,0.0f,1.0f));
        linesGeom->setColorArray(mycolors);
        linesGeom->setColorBinding(osg::Geometry::BIND_OVERALL);
        
        // set the normal in the same way color.
        osg::Vec3Array* normals = new osg::Vec3Array;
        normals->push_back(osg::Vec3(0.0f,-1.0f,0.0f));
        linesGeom->setNormalArray(normals);
        linesGeom->setNormalBinding(osg::Geometry::BIND_OVERALL);

        // This time we simply use primitive, and hardwire the number of coords to use 
        // since we know up front,
        linesGeom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::LINES,0,8));
             
        // add the points geomtry to the geode.
	
        myGeode->addDrawable(linesGeom);
}
// This routine sets upa color vector (RGBA) for use in shading the selected polygons
float  *returnColor(int terrainType)
{
	float *colorVec = new float[4];

	switch(terrainType) {
		case TWOINCHWASHBOARD:
			colorVec[0] = 1.0; 
			colorVec[1] = 1.0;
			colorVec[2] = 0.0;
			colorVec[3] = 1.0;
			break;
		case SIXINCHWASHBOARD:   
			colorVec[0] = 0.9; 
			colorVec[1] = 0.0;
			colorVec[2] = 0.75;
			colorVec[3] = 1.0;
			break;
		case RADIALWASHBOARD: 
			colorVec[0] = 0.75; 
			colorVec[1] = 0.4;
			colorVec[2] = 0.0;
			colorVec[3] = 1.0;
			break;
		case BELGIANBLOCK: 
			colorVec[0] = 1.0; 
			colorVec[1] = 0.0;
			colorVec[2] = 0.0;
			colorVec[3] = 1.0;
			break;
		case EMBEDDEDROCK:
			colorVec[0] = 0.0; 
			colorVec[1] = 0.0;
			colorVec[2] = 1.0;
			colorVec[3] = 1.0;
			break;
		case CROSSCOUNTRY:
			colorVec[0] = 0.0; 
			colorVec[1] = 1.0;
			colorVec[2] = 0.0;
			colorVec[3] = 1.0;
			break;
		case GRAVELROAD:
			colorVec[0] = 1,0;
			colorVec[1] = 1.0;
			colorVec[2] = 1.0;
			colorVec[3] = 1.0;
			break;
		case PAVEDROAD:
			colorVec[0] = 0.5; 
			colorVec[1] = 0.5;
			colorVec[2] = 0.5;
			colorVec[3] = 1.0;
			break;
		case NOTTRAVERSABLE:
			colorVec[0] = 0.1; 
			colorVec[1] = 0.1;
			colorVec[2] = 0.1;
			colorVec[3] = 1.0;
			break;
		default:
			colorVec[0] = 1.0; 
			colorVec[1] = 1.0;
			colorVec[2] = 0.0;
			colorVec[3] = 1.0;
			break;
	}

	return colorVec;

}

void DrawAreas(int polygonID, int iNumVertices)
{
		 // create Geometry object to store all the vetices and lines primtive.
        osg::Geometry* polyGeom = new osg::Geometry();
        
        // this time we'll a C arrays to initilize the vertices.
        // note, anticlockwsie ordering.
        // note II, OpenGL polygons must be convex plan polygons, otherwise 
        // undefined results will occur.  If you have concave polygons or ones
        // that cross over themselves then use the osgUtil::Tesselator to fix
        // the polygons into a set of valid polygons.
        osg::Vec3 myCoords[MAXVERTICES];

		int i = 0;
        // loop thru polygon map to find matching vertices
		std::multimap< int, osg::Vec3f >::iterator it =  mapPolygonPoints.begin();
		while( it !=  mapPolygonPoints.end() ) {
			if( it->first == polygonID ) {
				int LastKey = it->first;
				std::cout << it->first << " " << it->second << std::endl;
				myCoords[i] = it->second;
				i++;
				
			}
			++it;
		}

		//  loop thru color map to find desired color
		std::multimap< int, int >::iterator itc =  mapPolygonColor.begin();
		while( itc !=  mapPolygonColor.end() ) {
			if( itc->first == polygonID ) {
				selectedTerrainType = itc->second;	
			}
			++itc;
		}

	// set the colors as before, use a ref_ptr rather than just
    // standard C pointer, as that in the case of it not being
    // assigned it will still be cleaned up automatically.
    osg::ref_ptr<osg::Vec4Array> shared_colors = new osg::Vec4Array;
    float *colorVec = returnColor(selectedTerrainType);//new float[4];

    shared_colors->push_back(osg::Vec4(colorVec[0],colorVec[1],colorVec[2],colorVec[3]));//(osg::Vec4(1.0f,1.0f,0.0f,1.0f));

    // same trick for shared normal.
    osg::ref_ptr<osg::Vec3Array> shared_normals = new osg::Vec3Array;
    shared_normals->push_back(osg::Vec3(0.0f,-1.0f,0.0f));

        
//        int numCoords = sizeof(myCoords)/sizeof(osg::Vec3);
        
        //osg::Vec3Array* vertices2 = new osg::Vec3Array(numCoords,myCoords);
		osg::Vec3Array* vertices2 = new osg::Vec3Array(iNumVertices,myCoords);
       
        // pass the created vertex array to the points geometry object.
        polyGeom->setVertexArray(vertices2);
        
        // use the shared color array.
        polyGeom->setColorArray(shared_colors.get());
        polyGeom->setColorBinding(osg::Geometry::BIND_OVERALL);
        

        // use the shared normal array.
        polyGeom->setNormalArray(shared_normals.get());
        polyGeom->setNormalBinding(osg::Geometry::BIND_OVERALL);
        

        // This time we simply use primitive, and hardwire the number of coords to use 
        // since we know up front,
        polyGeom->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::POLYGON,0,iNumVertices));

        //printTriangles("Polygon",*polyGeom);
        
        // add the points geomtry to the geode.
        myGeode->addDrawable(polyGeom);

		// polygon stipple
        osg::StateSet* stateSet = new osg::StateSet();
        polyGeom->setStateSet(stateSet);
        osg::PolygonStipple* polygonStipple = new osg::PolygonStipple;
        stateSet->setAttributeAndModes(polygonStipple,osg::StateAttribute::OVERRIDE|osg::StateAttribute::ON);
}
// routine to "start" collecting polygon definition vertices
void StartPolygon(void)
{
	std::cout << "Start Polygon" << std::endl;
	if (!bInitFile) {
		std::string sPointFileName = (std::string(globalXMLFilename));
		
		stream.open (sPointFileName.c_str());
		stream << "<MGXML>"  << std::endl;
	    
		bInitFile = true;
	}

	stream << "<polygon>" <<  std::endl;
	stream << "<polyID>" << polygonCount << "</polyID>" << std::endl;
	stream << "<terraintype>" << selectedTerrainType << "</terrainType>" << std::endl;
	
	stream << "<datapoints>" <<  std::endl;
	vertexCount = 0;

	globalStartPolygon = false;
	

}
// routine to signal that all points to define a polygon have been specified
void StopPolygon(void)
{
	std::cout << "Stop Polygon" << std::endl;
	stream << "</datapoints>" <<  std::endl;
	stream << "</polygon>" <<  std::endl;
	polygonCount++;

	for (int i = 0; i < polygonCount; i++) {
		std::cout << "number of points in polygon #" << i << " is " << mapPolygonPoints.count(i) << std::endl;
		if (mapPolygonPoints.count(i) < 3) {
			MessageBox(NULL,"Polygons require at least 3 points to define!","Error!",MB_OK);
		}
		else
			DrawAreas( i,mapPolygonPoints.count(i) );
		
	}

	
	globalEndPolygon = false;
}


// class to handle events with a pick
class PickHandler : public osgGA::GUIEventHandler {
public: 

    PickHandler(osgText::Text* updateText):
        
        _updateText(updateText) {}
        
    ~PickHandler() {}
    
    bool handle(const osgGA::GUIEventAdapter& ea,osgGA::GUIActionAdapter& us);

    virtual void pick(const osgGA::GUIEventAdapter& ea);

    void setLabel(const std::string& name)
    {
        if (_updateText.get()) _updateText->setText(name);
    }
    
protected:

    
    osg::ref_ptr<osgText::Text>  _updateText;
};

bool PickHandler::handle(const osgGA::GUIEventAdapter& ea,osgGA::GUIActionAdapter&)
{
	
    switch(ea.getEventType())
    {
	case(osgGA::GUIEventAdapter::LEFT_MOUSE_BUTTON):
        {
            pick(ea);
        }
        return false;
        
    default:
        return false;
    }
}

void PickHandler::pick(const osgGA::GUIEventAdapter& ea)
{
		std::cout<<ea.getX()<<","<<ea.getY()<<" Mouse" << std::endl;   
		float newx, newy;
		newx = bb.xMin() + ((bb.xMax() - bb.xMin()) * (ea.getX() - XMOUSEMIN) / (XMOUSEMAX - XMOUSEMIN));
		newy = bb.yMin() + ((bb.yMax() - bb.yMin()) * (ea.getY() - YMOUSEMIN) / (YMOUSEMAX - YMOUSEMIN));
		std::cout<<newx<<","<<newy<<" World" << std::endl;  

		
        // write "XML" to output stream, points.txt
		stream << "<datafield count=\"" <<  vertexCount << " \" xval=\"" << newx+0.0 << 
			      " \" yval=\"" << newy+0.0 << "\" />" << std::endl;
		DrawBox(newx, newy);
		osg::Vec3f pickedPoint;
		pickedPoint.x() = newx;
		pickedPoint.y() = newy;
		pickedPoint.z() = 0.0;

		mapPolygonPoints.insert( std::pair<int,osg::Vec3f>(polygonCount, pickedPoint) );
		mapPolygonColor.insert( std::pair<int,int>(polygonCount, selectedTerrainType) );

		vertexCount++;
		  
}

Producer::CameraConfig* setupCameras()
{
   // set camera projection and lense params
   tankFollowerCamera->setProjectionRectangle(0.0f, 1.0f, 0.0f, 1.0f);
   tankFollowerCamera->setShareLens(false);
   tankFollowerCamera->getLens()->setAutoAspect(true);

   myCameraConfig->addCamera("tankFollowerCamera",tankFollowerCamera);
   
   // set up title bar and make sky black
   Producer::RenderSurface* rsOne = mainCamera->getRenderSurface();
   tankFollowerCamera->setRenderSurface( rsOne );
   tankFollowerCamera->setClearColor(0.0,0.0,0.0,1.0);
   rsOne->setWindowName("TA2 Terrain Mapper     Trideum Corp.");

   // set up window size and location
   rsOne->setWindowRectangle(800,200,640,640);
   return myCameraConfig;
}

void OSG_Module(void)
{
   // Group to serve as root of scene graph:
   osg::Group* root = NULL;
   // Node to contain geometry for a tank model:
   osg::Node* tankNode = NULL;
   // Node to contain geometry for terrain model:
   osg::Node* terrainNode = NULL;
   osg::Node* terrainNodeHUD = NULL;
   // Matrix transform node for positioning tank model:
   osg::PositionAttitudeTransform* tankXform;
   // Position of tank:
   osg::Vec3 tankPosit;
   // A geometry node for our HUD:
   osg::Geode* HUDGeode = new osg::Geode();
   // Text instance for HUD:
   osgText::Text* textOne = new osgText::Text();
   // Text for label that will follow the tank:
   osgText::Text* tankLabel = new osgText::Text();
   // Projection node for defining view frustrum for HUD:
   osg::Projection* HUDProjectionMatrix = new osg::Projection;

   osg::PositionAttitudeTransform* terrainXform;
   

   
   osg::Geometry* myGeometry = new osg::Geometry();

   // Initialize root of scene:
   root = new osg::Group();

   osgDB::FilePathList pathList = osgDB::getDataFilePathList();
   pathList.push_back("C:\\OSG1\\NPS_Tutorials\\NPSData\\Models\\T72-Tank\\");
   //pathList.push_back("C:\\OSG1\\NPS_Tutorials\\NPSData\\Models\\JoeDirt\\");
   for (int ii = 0 ; ii < globalNumberSearchPaths ; ii++) {
	   //MessageBox(NULL,globalSavedDir[ii].c_str(),"",MB_OK);
		pathList.push_back(globalSavedDir[ii]);
   }
   //pathList.push_back("C:\\OSG1\\NPS_Tutorials\\NPSData\\Textures\\");
   osgDB::setDataFilePathList(pathList);


   // Load models from files and assign to nodes:
   tankNode = osgDB::readNodeFile("t72-tank_des.flt");
   //terrainNode = osgDB::readNodeFile("JoeDirt.flt");
   terrainNode = osgDB::readNodeFile(globalOSGTerrainName);

   //terrainNodeHUD = osgDB::readNodeFile("JoeDirt.flt");
   terrainNodeHUD = osgDB::readNodeFile(globalOSGTerrainName);
   //MessageBox(NULL,globalOSGTerrainName.c_str(),"Terrain File to load",MB_OK);

   const osg::BoundingSphere bs = terrainNodeHUD->getBound();
   //osg::BoundingBox bb ; //= terrainNode->getBound();
   
   bb.expandBy(bs);
		std::cout << bb.xMin() << std::endl;
		std::cout << bb.xMax() << std::endl;
		std::cout << bb.yMin() << std::endl;
		std::cout << bb.yMax() << std::endl;
		std::cout << bb.zMin() << std::endl;
		std::cout << bb.zMax() << std::endl;

		boxScaleFactor = (int) ((bb.xMax() - bb.xMin()) / 500.0f) ;

	//sprintf(tempstring,"%f %f %f %f %f %f",bb.xMin(),bb.xMax(),bb.yMin(),bb.yMax(),bb.zMin(),bb.zMax());
	//MessageBox(NULL,tempstring,"",MB_OK);

   // Initialize transform to be used for positioning the tank
   // assign values for the transform
   tankXform = new osg::PositionAttitudeTransform();
   tankPosit.set(5,5,8);
   tankXform->setPosition( tankPosit );

   // Build the scene - add the terrain node directly to the root,
   // connect the tank node to the root via the transform node:
   // **** root->addChild(terrainNode);
   //root->addChild(tankXform);
   tankXform->addChild(tankNode);

   // setup Light source to act as Sun
   osg::Light * light = new osg::Light; 
   light->setLightNum (0); 

   // Set Ambient, Diffuse, And Specular values
    float sunAmbient = 0.5f; 
    float sunDiffuse = 0.6f; 
    float sunSpecular = 0.2f; 

    light->setDiffuse (osg::Vec4 (sunDiffuse, sunDiffuse, sunDiffuse, 1.0f)); 
    light->setAmbient (osg::Vec4 (sunAmbient, sunAmbient, sunAmbient, 1.0f)); 
    light->setSpecular (osg::Vec4 (sunSpecular, sunSpecular, sunSpecular, 1.0f)); 

    // Set Sun Position and direction
    light->setPosition (osg::Vec4(0.0f, 0.0f, 1000.0f, 0.0f)); 
    light->setDirection (osg::Vec3 (0.0f, 0.0f, 0.0f)); 

    // Add to scene tree
    osg::LightSource* sun = new osg::LightSource; 
    sun->setLight (light); 
    root->addChild (sun); 

	Producer::CameraConfig* myCameraConfig = setupCameras();
    osgProducer::Viewer viewer(myCameraConfig);
    viewer.setUpViewer(osgProducer::Viewer::STANDARD_SETTINGS);
  

   // Initialize the projection matrix for viewing everything we
   // will add as descendents of this node. Use screen coordinates
   // to define the horizontal and vertical extent of the projection
   // matrix. Positions described under this node will equate to
   // pixel coordinates.
   //mg HUDProjectionMatrix->setMatrix(osg::Matrix::ortho2D(0,1024,0,768));

   

   xdiff = (bb.xMax() + bb.xMin()) * 0.5;
   ydiff = (bb.yMax() + bb.yMin()) * 0.5;
   zdiff = (bb.zMax() + bb.zMin()) * 0.5;

   
   xoff = bb.xMin() - xdiff;
   yoff = bb.yMin() - ydiff;
   zoff = bb.zMin() - zdiff;

   HUDProjectionMatrix->setMatrix(osg::Matrix::ortho2D((int) bb.xMin(),
													   (int) bb.xMax(),
													   (int) bb.yMin(),
													   (int) bb.yMax()));
  
   // For the HUD model view matrix use an identity matrix:
   osg::MatrixTransform* HUDModelViewMatrix = new osg::MatrixTransform;
   HUDModelViewMatrix->setMatrix(osg::Matrix::identity());

   // Make sure the model view matrix is not affected by any transforms
   // above it in the scene graph:
   HUDModelViewMatrix->setReferenceFrame(osg::Transform::ABSOLUTE_RF);

   // Add the HUD projection matrix as a child of the root node
   // and the HUD model view matrix as a child of the projection matrix
   // Anything under this node will be view using this projection matrix
   // and positioned with this model view matrix.
   root->addChild(HUDProjectionMatrix);
   HUDProjectionMatrix->addChild(HUDModelViewMatrix);
   HUDProjectionMatrix->addChild(HUDModelViewMatrix);

   // Add the Geometry node to contain HUD geometry as a child of the
   // HUD model view matrix.
   // (See figure "n")
   //HUDModelViewMatrix->addChild( HUDGeode );

   terrainXform = new osg::PositionAttitudeTransform();
   
   terrainXform->setPosition( osg::Vec3(-xdiff*0.00,
	                                    -ydiff*0.00,
										-zdiff*1.0) );
   terrainXform->setAttitude( osg::Quat(osg::DegreesToRadians(00.0), osg::Vec3(1,0,0)) );
   
   HUDModelViewMatrix->addChild(terrainXform);
   terrainXform->addChild(terrainNodeHUD);
   terrainXform->addChild(myGeode);
   //HUDModelViewMatrix->addChild(terrainNodeHUD);
   //HUDModelViewMatrix->addChild(myGeode);

   osg::StateSet* myHUDStateSet = new osg::StateSet();
   myGeode->setStateSet(myHUDStateSet);

   // Disable depth testing so geometry is draw regardless of depth values
   // of geometry already draw.
   myHUDStateSet->setMode(GL_DEPTH_TEST,osg::StateAttribute::OFF);
  // myHUDStateSet->setRenderingHint( osg::StateSet::TRANSPARENT_BIN );

   // Need to make sure this geometry is draw last. RenderBins are handled
   // in numerical order so set bin number to 11
   myHUDStateSet->setRenderBinDetails( 98, "RenderBin");
   

   // Add the text (Text class is derived from drawable) to the geode:
   HUDGeode->addDrawable( textOne );

   // Set up the parameters for the text we'll add to the HUD:
   textOne->setCharacterSize(25);
   textOne->setFont("C:/WINDOWS/Fonts/impact.ttf");
   textOne->setText("Not so good");
   textOne->setAxisAlignment(osgText::Text::SCREEN);
   textOne->setPosition( osg::Vec3(360,165,-1.5) );
   textOne->setColor( osg::Vec4(199, 77, 15, 1) );

   // Set up the parameters for the text label for the tank
   // align text with tank's SCREEN.
   // (for Onder: use XZ_PLANE to align text with tank's XZ plane.)
   tankLabel->setCharacterSize(5);
   tankLabel->setFont("/fonts/arial.ttf");
   tankLabel->setText("Tank #1");
   tankLabel->setAxisAlignment(osgText::Text::XZ_PLANE);

   // Set the text to render with alignment anchor and bounding box around it:
   tankLabel->setDrawMode(osgText::Text::TEXT |
      osgText::Text::ALIGNMENT |
      osgText::Text::BOUNDINGBOX);
   tankLabel->setAlignment(osgText::Text::CENTER_TOP);
   tankLabel->setPosition( osg::Vec3(0,0,8) );
   tankLabel->setColor( osg::Vec4(1.0f, 0.0f, 0.0f, 1.0f) );

   // Declare a geode to contain the tank's text label:
   osg::Geode* tankLabelGeode = new osg::Geode();

   // Add the tank label to the scene using a geode:
   tankLabelGeode->addDrawable(tankLabel);
   tankXform->addChild(tankLabelGeode);

   // Set up geometry for the HUD and add it to the HUD
   // geometry node, see tutorial "n" for details:
   osg::Geometry* HUDBackgroundGeometry = new osg::Geometry();

   osg::Vec3Array* HUDBackgroundVertices = new osg::Vec3Array;

   HUDBackgroundVertices->push_back( osg::Vec3( 0, 0,-1) );
   HUDBackgroundVertices->push_back( osg::Vec3(1024, 0,-1) );
   HUDBackgroundVertices->push_back( osg::Vec3(1024,768,-1) );
   HUDBackgroundVertices->push_back( osg::Vec3( 0,768,-1) );

   osg::DrawElementsUInt* HUDBackgroundIndices =
      new osg::DrawElementsUInt(osg::PrimitiveSet::POLYGON, 0);
   HUDBackgroundIndices->push_back(0);
   HUDBackgroundIndices->push_back(1);
   HUDBackgroundIndices->push_back(2);
   HUDBackgroundIndices->push_back(3);

   osg::Vec4Array* HUDcolors = new osg::Vec4Array;
   HUDcolors->push_back(osg::Vec4(0.8f,0.8f,0.8f,0.8f));

   osg::Vec2Array* texcoords = new osg::Vec2Array(4);
   (*texcoords)[0].set(0.0f,0.0f);
   (*texcoords)[1].set(1.0f,0.0f);
   (*texcoords)[2].set(1.0f,1.0f);
   (*texcoords)[3].set(0.0f,1.0f);
   HUDBackgroundGeometry->setTexCoordArray(0,texcoords);
   osg::Texture2D* HUDTexture = new osg::Texture2D;
   HUDTexture->setDataVariance(osg::Object::DYNAMIC);
   osg::Image* hudImage;
   hudImage = osgDB::readImageFile("HUDBack2.tga");
   HUDTexture->setImage(hudImage);
   osg::Vec3Array* HUDnormals = new osg::Vec3Array;
   HUDnormals->push_back(osg::Vec3(0.0f,0.0f,1.0f));
   HUDBackgroundGeometry->setNormalArray(HUDnormals);
   HUDBackgroundGeometry->setNormalBinding(osg::Geometry::BIND_OVERALL);

   HUDBackgroundGeometry->addPrimitiveSet(HUDBackgroundIndices);
   HUDBackgroundGeometry->setVertexArray(HUDBackgroundVertices);
   HUDBackgroundGeometry->setColorArray(HUDcolors);
   HUDBackgroundGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);

   HUDGeode->addDrawable(HUDBackgroundGeometry);
   // Create and set up a state set using the texture from above:
   osg::StateSet* HUDStateSet = new osg::StateSet();
   HUDGeode->setStateSet(HUDStateSet);
   HUDStateSet->
      setTextureAttributeAndModes(0,HUDTexture,osg::StateAttribute::ON);

   // For this state set, turn blending on (so alpha texture looks right)
   HUDStateSet->setMode(GL_BLEND,osg::StateAttribute::ON);

   // Disable depth testing so geometry is draw regardless of depth values
   // of geometry already draw.
   HUDStateSet->setMode(GL_DEPTH_TEST,osg::StateAttribute::OFF);
   HUDStateSet->setRenderingHint( osg::StateSet::TRANSPARENT_BIN );

   // Need to make sure this geometry is draw last. RenderBins are handled
   // in numerical order so set bin number to 11
   HUDStateSet->setRenderBinDetails( 99, "RenderBin");

   osg::ref_ptr<osgText::Text> updateText = new osgText::Text;
   // add the handler for doing the picking
    viewer.getEventHandlerList().push_front(new PickHandler(updateText.get()));

   keyboardEventHandler* key_b = new keyboardEventHandler();
   key_b->addFunction('b',StartPolygon);
   viewer.getEventHandlerList().push_front(key_b);

   keyboardEventHandler* key_e = new keyboardEventHandler();
   key_e->addFunction('e',StopPolygon);
   viewer.getEventHandlerList().push_front(key_e);

   keyboardEventHandler* key_1 = new keyboardEventHandler();
   key_1->addFunction('1',TerrainSelectTWOINCHWASHBOARD);
   viewer.getEventHandlerList().push_front(key_1);

   keyboardEventHandler* key_2 = new keyboardEventHandler();
   key_2->addFunction('2',TerrainSelectSIXINCHWASHBOARD);
   viewer.getEventHandlerList().push_front(key_2);

   keyboardEventHandler* key_3 = new keyboardEventHandler();
   key_3->addFunction('3',TerrainSelectRADIALWASHBOARD);
   viewer.getEventHandlerList().push_front(key_3);

   keyboardEventHandler* key_4 = new keyboardEventHandler();
   key_4->addFunction('4',TerrainSelectBELGIANBLOCK);
   viewer.getEventHandlerList().push_front(key_4);

   keyboardEventHandler* key_5 = new keyboardEventHandler();
   key_5->addFunction('5',TerrainSelectEMBEDDEDROCK);
   viewer.getEventHandlerList().push_front(key_5);

   keyboardEventHandler* key_6 = new keyboardEventHandler();
   key_6->addFunction('6',TerrainSelectCROSSCOUNTRY);
   viewer.getEventHandlerList().push_front(key_6);

   keyboardEventHandler* key_7 = new keyboardEventHandler();
   key_7->addFunction('7',TerrainSelectGRAVELROAD);
   viewer.getEventHandlerList().push_front(key_7);

   keyboardEventHandler* key_8 = new keyboardEventHandler();
   key_8->addFunction('8',TerrainSelectPAVEDROAD);
   viewer.getEventHandlerList().push_front(key_8);

   keyboardEventHandler* key_9 = new keyboardEventHandler();
   key_9->addFunction('9',TerrainSelectNOTTRAVERSABLE);
   viewer.getEventHandlerList().push_front(key_9);

   keyboardEventHandler* key_x = new keyboardEventHandler();
   key_x->addFunction('x',ExitApp);
   viewer.getEventHandlerList().push_front(key_x);

   
   //MessageBox(NULL,globalOSGTerrainName.c_str(),"",MB_OK);
   viewer.setSceneData( root );

   viewer.realize();

   while( !viewer.done() )
   {
      viewer.sync();
      viewer.update();

	  if (globalStartPolygon) StartPolygon();

	  if (globalEndPolygon)   StopPolygon();

      viewer.frame();
   }
   return ;
} 

